(* Copyright (c) 2011 - 2015. Cary Jensen, Jensen Data Systems, Inc.
   This code sample is intended for the readers of
   "Delphi in Depth: ClientDataSets"
     and
   "Delphi in Depth: ClientDataSets 2nd Edition"
   by Cary Jensen

   No guarantees or warranties are expressed or implied concerning
   the applicability of techniques or code included in this example
   or in the accompanying book. If you wish to use techniques or
   code included in this example or described in the book, it is
   your responsibility to test and certify any code, techniques,
   or design you adopt.

   For information on consulting or training services, please visit:
   http://www.JensenDataSystems.com

   For more information about "Delphi in Depth: ClientDataSets 2nd Edition"
   by Cary Jensen, including links to order the book, please visit:
   http://www.JensenDataSystems.com/cdsbook2
*)
unit mainformu;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, DBCtrls, Grids, DBGrids, DB, ExtCtrls, DBClient;

type
  TForm1 = class(TForm)
    Panel1: TPanel;
    Panel2: TPanel;
    Panel3: TPanel;
    Panel4: TPanel;
    Splitter1: TSplitter;
    Splitter2: TSplitter;
    DataSource1: TDataSource;
    DataSource2: TDataSource;
    DataSource3: TDataSource;
    DBGrid1: TDBGrid;
    DBGrid2: TDBGrid;
    DBGrid3: TDBGrid;
    DBNavigator1: TDBNavigator;
    DBNavigator2: TDBNavigator;
    DBNavigator3: TDBNavigator;
    CreateButton: TButton;
    ClearButton: TButton;
    SaveButton: TButton;
    CheckBox1: TCheckBox;
    LoadButton: TButton;
    OpenDialog1: TOpenDialog;
    procedure CreateButtonClick(Sender: TObject);
    procedure ClearButtonClick(Sender: TObject);
    procedure SaveButtonClick(Sender: TObject);
    procedure LoadButtonClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    TopLevelCDS: TClientDataSet;
    MidLevelCDS: TClientDataSet;
    ThirdLevelCDS: TClientDataSet;
    procedure CreateNestedDatasets;
    procedure CleanUpNestedDataSets;
  end;

var
  Form1: TForm1;

const
  cdsfile = 'data.cds';
  xmlfile = 'data.xml';

implementation

{$R *.dfm}

procedure TForm1.ClearButtonClick(Sender: TObject);
begin
  ClearButton.Enabled := False;
  CreateButton.Enabled := True;
  SaveButton.Enabled := False;
  CleanUpNestedDataSets;
end;

procedure TForm1.CreateButtonClick(Sender: TObject);
begin
  ClearButton.Enabled := True;
  CreateButton.Enabled := False;
  SaveButton.Enabled := True;
  CreateNestedDatasets;
end;

procedure TForm1.CreateNestedDataSets;
begin

  TopLevelCDS := TClientDataSet.Create(Self);
  MidLevelCDS := TClientDataSet.Create(TopLevelCDS);
  ThirdLevelCDS := TClientDataSet.Create(TopLevelCDS);

  with TIntegerField.Create(Self) do
  begin
    Name := 'TopID';
    FieldKind := fkData;
    FieldName := 'ID';
    DataSet := TopLevelCDS;
    Required := True;
  end;
  with TStringField.Create(Self) do
  begin
    Name := 'TopName';
    FieldKind := fkData;
    FieldName := 'Name';
    Size := 40;
    DataSet := TopLevelCDS;
  end;
  with TMemoField.Create(Self) do
  begin
    Name := 'TopComments';
    FieldKind := fkData;
    FieldName := 'Comments';
    DataSet := TopLevelCDS;
  end;
  with TDateField.Create(Self) do
  begin
    Name := 'TopDateInitiated';
    FieldKind := fkData;
    FieldName := 'Date Initiated';
    DataSet := TopLevelCDS;
  end;
  //Note: For TDataSetFields, FieldKind is fkDataSet by default
  with TDataSetField.Create(Self) do
  begin
    Name := 'TopNested';
    FieldName := 'NestedDataSet';
    DataSet := TopLevelCDS;
  end;

  //MideLevelCDS
  MidLevelCDS.DataSetField := TDataSetField(FindComponent('TopNested'));
  with TIntegerField.Create(Self) do
  begin
    Name := 'MidID';
    FieldKind := fkData;
    FieldName := 'MidID';
    DataSet := MidLevelCDS;
    Required := True;
  end;
  with TStringField.Create(Self) do
  begin
    Name := 'MidName';
    FieldKind := fkData;
    FieldName := 'MidName';
    DataSet := MidLevelCDS;
    Size := 30;;
  end;
  with TDataSetField.Create(Self) do
  begin
    Name := 'MidNested';
    FieldName := 'NestedNestedDataSet';
    DataSet := MidLevelCDS;
  end;

  //Third Level
  ThirdLevelCDS.DataSetField := TDataSetField(FindComponent('MidNested'));
  with TIntegerField.Create(Self) do
  begin
    Name := 'ThirdID';
    FieldKind := fkData;
    FieldName := 'ThirdID';
    DataSet := ThirdLevelCDS;
    Required := True;
  end;
  with TStringField.Create(Self) do
  begin
    Name := 'ThirdName';
    FieldKind := fkData;
    FieldName := 'ThirdName';
    DataSet := ThirdLevelCDS;
    Size := 25;
  end;
  with TBooleanField.Create(Self) do
  begin
    Name := 'ThirdActive';
    FieldKind := fkData;
    FieldName := 'ThirdActive';
    DataSet := ThirdLevelCDS;
  end;
  //Create the ClientDataSet and its nested datasets
  TopLevelCDS.CreateDataSet;
  //Configure the DataSources
  DataSource1.DataSet := TopLevelCDS;
  DataSource2.DataSet := MidLevelCDS;
  Datasource3.DataSet := ThirdLevelCDS;
end;

procedure TForm1.SaveButtonClick(Sender: TObject);
begin
  if CheckBox1.Checked then
    TopLevelCDS.SaveToFile(xmlfile, dfXML)
  else
  begin
    //This will not be able to load properly
    TopLevelCDS.SaveToFile(cdsfile, dfBinary);
  end;
end;

procedure TForm1.LoadButtonClick(Sender: TObject);
begin
if OpenDialog1.Execute then
begin
  if TopLevelCDS <> nil then
    CleanUpNestedDataSets;
  TopLevelCDS := TClientDataSet.Create(Self);
  TopLevelCDS.LoadFromFile(OpenDialog1.Filename);
  DataSource1.DataSet := TopLevelCDS;
  (**)
  MidLevelCDS := TClientDataSet.Create(TopLevelCDS);
  MidLevelCDS.DataSetField := TDataSetField(TopLevelCDS.FieldByName('NestedDataSet'));
  ThirdLevelCDS := TClientDataSet.Create(TopLevelCDS);
  ThirdLevelCDS.DataSetField := TDataSetField(MidLevelCDS.FieldByName('NestedNestedDataSet'));
  DataSource2.DataSet := MidLevelCDS;
  DataSource3.DataSet := ThirdLevelCDS;
  (**)
  (*
  DataSource2.DataSet :=
    TDataSetField(TopLevelCDS.FieldByName('NestedDataSet')).NestedDataSet;
  DataSource3.DataSet :=
    TDataSetfield(DataSource2.DataSet.FieldByName('NestedNestedDataSet')).NestedDataSet;
  (**)  ClearButton.Enabled := True;
  SaveButton.Enabled := True;
end;
end;

procedure TForm1.CleanUpNestedDataSets;
begin
  DataSource1.DataSet := nil;
  TopLevelCDS.Close;
  FreeAndNil(TopLevelCDS);
end;

end.
